/*
 * File: cell.cpp
 *
 * Description: The "Duality" screensaver.
 *
 * Author: Mark Wagner
 *
 * Copyright (c) 2003 Mark Wagner
 */

/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include <windows.h>
#include <scrnsave.h>
#include <stdio.h>
#include <stdlib.h>

#include "resource.h"


TCHAR szAppName[] = "ScreenSaver.Duality";
const char IniFile[] = "control.ini";


typedef struct
{
    int x;
    int y;
    int type;
} Cell;

int *board;
Cell *cells;
int NumCells;
int ScreenWidth;
int ScreenHeight;
int width;
int height;

float InitialFill = 10;   // Initial fill ratio
int   Delay = (1000/100);   // Frame delay in milleseconds
int   PixelSize = 2;        // Pixel size.  May look better if it's a power of 2
float Bias = 0;             // Bias between between red and blue.  Bias of -1 is pure red/pink, bias of 1 is pure blue/cyan

UINT timer = 0;
bool PrecisionTimerUsed = false;
COLORREF colors[4] = {0x00FFCF80, 0x00FF8040, 0x00CF80FF, 0x008040FF};
HWND MainWindow;
HDC BackBuffer;
HBITMAP BufferBMP;
HBITMAP OldBMP;
extern HINSTANCE hMainInstance;

void MoveCell(int index);
int getboard(int x, int y);
void CALLBACK IterateCallback(UINT uID, UINT msg, DWORD dw1, DWORD dw2, DWORD dw3);
void LoadPrefs();
void SavePrefs();

LRESULT WINAPI ScreenSaverProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    int i;
    int TrueNumCells;
    int x, y, type;
    HDC hdc;
    RECT rc;
    TIMECAPS tc;

    switch(msg)
    {
        case WM_CREATE:
            // Load prefs
            LoadPrefs();

            // Get initial variables
            MainWindow = hWnd;
            
            GetClientRect(hWnd, &rc);
            ScreenWidth = rc.right - rc.left;
            ScreenHeight = rc.bottom - rc.top;
            width = ScreenWidth / PixelSize;
            height = ScreenHeight / PixelSize;
            NumCells = width * height;

            // Allocate the back buffer
            hdc = GetDC(hWnd);
            BackBuffer = CreateCompatibleDC(hdc);
            BufferBMP = CreateCompatibleBitmap(hdc, width, height);
            ReleaseDC(hWnd, hdc);
            OldBMP = (HBITMAP)SelectObject(BackBuffer, BufferBMP);

            // Allocate memory
            board = new int[NumCells];
            memset(board, 0, NumCells*sizeof(int));
            NumCells *= (InitialFill/100.0);
            cells = new Cell[NumCells];

            // Create the initial fill
            TrueNumCells = 0;
            for(i = 0; i < NumCells; i++)
            {
                x = rand() % width;
                y = rand() % height;
                type = (rand() % 4) + 1;
                if(getboard(x, y) == 0)
                {
                    board[(y*width)+x] = type;
                    cells[TrueNumCells].type = type;
                    cells[TrueNumCells].x = x;
                    cells[TrueNumCells].y = y;
                    TrueNumCells++;
                }
            }
            NumCells = TrueNumCells;

            // If possible, create a high-resolution timer
            if (timeGetDevCaps(&tc, sizeof(TIMECAPS)) == TIMERR_NOERROR) 
            {
                Delay = min(max(tc.wPeriodMin, Delay), tc.wPeriodMax);
                //timeBeginPeriod(wTimerRes);
                timer = timeSetEvent(Delay, Delay/2, IterateCallback, 1, TIME_ONESHOT);

                PrecisionTimerUsed = true;
            }
            else // Error.  Create standard timer            
            {
                timer = SetTimer(hWnd, 1, Delay, NULL);
            }
            break;
        case WM_DESTROY:
            delete [] board;
            delete [] cells;
            // Delete the timer
            if(PrecisionTimerUsed)
            {
                if(timer != 0)
                {
                    timeKillEvent(timer);
                    timer = 0;
                }
            }
            else
            {
                KillTimer(hWnd, timer);
            }
            SelectObject(BackBuffer, OldBMP);
            DeleteObject(BufferBMP);
            DeleteDC(BackBuffer);
            break;

        case WM_TIMER:
            IterateCallback(0,0,0,0,0);
            break;
    }
    return DefScreenSaverProc(hWnd, msg, wParam, lParam);
}

BOOL WINAPI RegisterDialogClasses(HANDLE hInst)
{
    return TRUE;
}

void UpdateDisplay(HWND hDlg)
{
    char buffer[100];
    sprintf(buffer, "%d%%", (int)InitialFill);
    SetWindowText(GetDlgItem(hDlg, LBL_RATIO), buffer);

    sprintf(buffer, "%d ms", Delay);
    SetWindowText(GetDlgItem(hDlg, LBL_ANIDELAY), buffer);

    sprintf(buffer, "%d", PixelSize);
    SetWindowText(GetDlgItem(hDlg, LBL_PIXELSIZE), buffer);
}


BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static HWND scroll_anidelay;
    static HWND scroll_pixelsize;
    static HWND scroll_initialfill;

    switch(msg)
    {
        case WM_INITDIALOG:
            LoadPrefs();
            // Update the controls
            scroll_anidelay = GetDlgItem(hDlg, SCROLL_ANIDELAY);
            SetScrollRange(scroll_anidelay, SB_CTL, 1, 100, FALSE);
            SetScrollPos(scroll_anidelay, SB_CTL, Delay, TRUE);

            scroll_initialfill = GetDlgItem(hDlg, SCROLL_RATIO);
            SetScrollRange(scroll_initialfill, SB_CTL, 1, 20, FALSE);
            SetScrollPos(scroll_initialfill, SB_CTL, InitialFill, TRUE);

            scroll_pixelsize = GetDlgItem(hDlg, SCROLL_PIXELSIZE);
            SetScrollRange(scroll_pixelsize, SB_CTL, 1, 32, FALSE);
            SetScrollPos(scroll_pixelsize, SB_CTL, PixelSize, TRUE);

            UpdateDisplay(hDlg);
            return TRUE;
        case WM_HSCROLL:
            if((HWND)lParam == scroll_initialfill)
            {
                switch(LOWORD(wParam))
                {
                    case SB_RIGHT:
                        InitialFill = 1;
                        break;
                    case SB_LEFT:
                        InitialFill = 20;
                        break;
                    case SB_LINELEFT:
                    case SB_PAGELEFT:
                        InitialFill--;
                        break;
                    case SB_LINERIGHT:
                    case SB_PAGERIGHT:
                        InitialFill++;
                        break;
                    case SB_THUMBPOSITION:
                    case SB_THUMBTRACK:
                        InitialFill = HIWORD(wParam);
                        break;
                    case SB_ENDSCROLL:
                        return TRUE;
                }
                if(InitialFill > 20) InitialFill = 20;
                if(InitialFill < 0) InitialFill = 0;
                SetScrollPos(scroll_initialfill, SB_CTL, InitialFill, TRUE);
            }
            else if((HWND)lParam == scroll_pixelsize)
            {
                switch(LOWORD(wParam))
                {
                    case SB_RIGHT:
                        PixelSize = 1;
                        break;
                    case SB_LEFT:
                        PixelSize = 20;
                        break;
                    case SB_LINELEFT:
                        PixelSize--;
                        break;
                    case SB_PAGELEFT:
                        PixelSize /= 2;
                        break;
                    case SB_LINERIGHT:
                        PixelSize++;
                        break;
                    case SB_PAGERIGHT:
                        PixelSize *= 2;
                        break;
                    case SB_THUMBPOSITION:
                    case SB_THUMBTRACK:
                        PixelSize = HIWORD(wParam);
                        break;
                    case SB_ENDSCROLL:
                        return TRUE;
                }
                if(PixelSize > 32) PixelSize = 32;
                if(PixelSize < 1) PixelSize = 1;
                SetScrollPos(scroll_pixelsize, SB_CTL, PixelSize, TRUE);
            }
            else if((HWND)lParam == scroll_anidelay)
            {
                switch(LOWORD(wParam))
                {
                    case SB_RIGHT:
                        Delay = 1;
                        break;
                    case SB_LEFT:
                        Delay = 100;
                        break;
                    case SB_LINELEFT:
                        Delay--;
                        break;
                    case SB_PAGELEFT:
                        Delay -= 10;
                        break;
                    case SB_LINERIGHT:
                        Delay++;
                        break;
                    case SB_PAGERIGHT:
                        Delay += 10;
                        break;
                    case SB_THUMBPOSITION:
                    case SB_THUMBTRACK:
                        Delay = HIWORD(wParam);
                        break;
                    case SB_ENDSCROLL:
                        return TRUE;
                }
                if(Delay > 100) Delay = 100;
                if(Delay < 1) Delay = 1;
                SetScrollPos(scroll_anidelay, SB_CTL, Delay, TRUE);
            }
            UpdateDisplay(hDlg);
            return 0;
        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
                case IDOK:
                    SavePrefs();
                    EndDialog(hDlg, IDOK);
                case IDCANCEL:
                    EndDialog(hDlg, IDCANCEL);
            }
            break;
    }
    return FALSE;
}

void CALLBACK IterateCallback(UINT uID, UINT msg, DWORD precision, DWORD dw2, DWORD dw3)
{
    int i;
    HDC hdc;
    RECT rc;
    
    if(precision)
        timer = 0;

    // Update the cells
    for(i = 0; i < NumCells; i++)
    {
        MoveCell(i);
    }
    // Clear the screen
    rc.top = rc.left = 0;
    rc.bottom = height;
    rc.right = width;
    FillRect (BackBuffer, &rc, (HBRUSH)GetStockObject(BLACK_BRUSH)); 
    // Draw
    for(i = 0; i < NumCells; i++)
    {
        SetPixel(BackBuffer, cells[i].x, cells[i].y, colors[cells[i].type-1]);
    }

    // Swap buffers
    hdc = GetDC(MainWindow);
    if(PixelSize != 0)
    {
        StretchBlt(hdc, 0, 0, ScreenWidth, ScreenHeight, BackBuffer, 0, 0, width, height, SRCCOPY);
    }
    else
    {
        BitBlt(hdc, 0, 0, ScreenWidth, ScreenHeight, BackBuffer, 0, 0, SRCCOPY);
    }
    ReleaseDC(MainWindow, hdc);
    if(1)
    {
        // Precision timer is used.  Set up for another call
        timer = timeSetEvent(Delay, Delay/2, IterateCallback, 1, TIME_ONESHOT);
    }
}

void LoadPrefs()
{
    Delay = GetPrivateProfileInt(szAppName, "AnimDelay", 25, IniFile);
    PixelSize = GetPrivateProfileInt(szAppName, "PixelSize", 2, IniFile);
    InitialFill = GetPrivateProfileInt(szAppName, "FillRatio", 10, IniFile);

    // Inject sanity
    if(Delay < 0 || Delay > 1000)
        Delay = 25;
    if(PixelSize < 0)
        PixelSize = 1;
    if(InitialFill <= 0)
        InitialFill = 1;
    if(InitialFill > 100)
        InitialFill = 100; 
}

void SavePrefs()
{
    char buffer[100];
    sprintf(buffer, "%d", Delay);
    WritePrivateProfileString(szAppName, "AnimDelay", buffer, IniFile);
    
    sprintf(buffer, "%d", PixelSize);
    WritePrivateProfileString(szAppName, "PixelSize", buffer, IniFile);

    sprintf(buffer, "%d", (int)InitialFill);
    WritePrivateProfileString(szAppName, "FillRatio", buffer, IniFile);
}

int getboard(int x, int y)
{
    if(x >= width) x -= width;
    if(x < 0) x += width;
    if(y >= height) y -= height;
    if(y < 0) y += height;
    return board[(y*width)+x];
}

void MoveCell(int index)
{
    int x, y;
    int dx = 1, dy = 1;
    x = cells[index].x; y = cells[index].y;

    switch(cells[index].type)
    {
        case 1:
            if(getboard(x-1,y+1) == 0 && getboard(x+1,y+1) == 0 && getboard(x,y+1) == 0) 
                {
                    dx = 0; dy = 1;
                }
            else if(getboard(x-1,y) == 2 && getboard(x-1,y-1) != 0)
                {
                    dx = 0; dy = -1;
                }
            else if(getboard(x-1,y) == 2 && getboard(x-1,y-1) == 0)
                {
                    dx = -1; dy = 1;
                }
            else if(getboard(x+1,y) == 1 && getboard(x+1,y-1) != 0)
                {
                    dx = 0; dy = -1;
                }
            else if(getboard(x+1,y) == 1 && getboard(x+1,y-1) == 0)
                {
                    dx = 1; dy = -1;
                }
            else
            {
                dx = (rand() % 3) - 1; dy = 0;
            }
            break;
        case 2:
            if(getboard(x-1,y+1) == 0 && getboard(x+1,y+1) == 0 && getboard(x,y+1) == 0) 
                {
                    dx = 0; dy = 1;
                }
            else if(getboard(x+1,y) == 1 && getboard(x+1,y-1) != 0)
                {
                    dx = 0; dy = -1;
                }
            else if(getboard(x+1,y) == 1 && getboard(x+1,y-1) == 0)
                {
                    dx = 1; dy = -1;
                }
            else if(getboard(x-1,y) == 2 && getboard(x-1,y-1) != 0)
                {
                    dx = 0; dy = -1;
                }
            else if(getboard(x-1,y) == 2 && getboard(x-1,y-1) == 0)
                {
                    dx = -1; dy = 1;
                }
            else
            {
                dx = (rand() % 3) - 1; dy = 0;
            }
            break;
        case 3:
            if(getboard(x-1,y-1) == 0 && getboard(x+1,y-1) == 0 && getboard(x,y-1) == 0) 
                {
                    dx = 0; dy = -1;
                }
            else if(getboard(x-1,y) == 4 && getboard(x-1,y+1) != 0)
                {
                    dx = 0, dy = 1;
                }
            else if(getboard(x-1,y) == 4 && getboard(x-1,y+1) == 0)
                {
                    dx = -1, dy = 1;
                }
            else if(getboard(x+1,y) == 3 && getboard(x+1,y+1) != 0)
                {
                    dx = 0, dy = 1;
                }
            else if(getboard(x+1,y) == 3 && getboard(x+1,y+1) == 0)
                {
                    dx = 1, dy = 1;
                }
            else
            {
                dx = (rand() % 3) - 1; dy = 0;
            }
            break;
        case 4:
            if(getboard(x-1,y-1) == 0 && getboard(x+1,y-1) == 0 && getboard(x,y-1) == 0) 
                {
                    dx = 0; dy = -1;
                }
            else if(getboard(x+1,y) == 3 && getboard(x+1,y+1) != 0)
                {
                    dx = 0, dy = 1;
                }
            else if(getboard(x+1,y) == 3 && getboard(x+1,y+1) == 0)
                {
                    dx = 1, dy = 1;
                }
            else if(getboard(x-1,y) == 4 && getboard(x-1,y+1) != 0)
                {
                    dx = 0, dy = 1;
                }
            else if(getboard(x-1,y) == 4 && getboard(x-1,y+1) == 0)
                {
                    dx = -1, dy = 1;
                }
            else
            {
                dx = (rand() % 3) - 1; dy = 0;
            }
            break;
        default:
            MessageBeep(MB_ICONHAND);
    }
    if(getboard(x+dx,y+dy) == 0)
    {
        board[(y*width)+x] = 0;
        x += dx; y += dy;
        if(x >= width) x -= width;
        if(x < 0) x += width;
        if(y >= height) y -= height;
        if(y < 0) y += height;
        cells[index].x = x; cells[index].y = y;
        board[(y*width)+x] = cells[index].type;

    }
}
